#!/usr/bin/perl -w

# qmail-expand - show where an address would be delivered
# Copyright 1997 by Thomas Erskine under the GNU copyleft.

# - - -   Configuration   - - -

# Where is qmail?
$qmaildir = '/var/qmail';
# What separates the extension from the local part?
$dash = '-';
# What is the alias user called?
$alias = 'alias';

# - - -   Setup   - - -

# Which version is this?  In case I actually release it.
$version = '1.1';

# Version history:
# 1.1	- [TEE] fix virtual domains
# 1.0	- [TEE] initial release

# to resolve remote host MX records
if (&find_in_inc("Net/DNS.pm")) { $DNS_available = 1; }
else { $DNS_available = 0; }

# Parse the command-line
use Getopt::Std;
getopts('a:de:hq:');

if (defined $opt_h) { &usage; } # no return
if (defined $opt_a) { $alias = $opt_a; }
if (defined $opt_d) { $debug = 1; } else { $debug = 0; }
if (defined $opt_e) { $dash = $opt_e; }
if (defined $opt_q) { $qmaildir = $opt_q; }

unless ($#ARGV >= 0) { &usage; } # no return

# The various files we need to look in
$control = "$qmaildir/control";
$locals = "$control/locals";
$me = "$control/me";
$virtualdomains = "$control/virtualdomains";
$recipientmap = "$control/recipientmap";
$defaulthost = "$control/defaulthost";
$defaultdomain = "$control/defaultdomain";
$plusdomain = "$control/plusdomain";
$smtproutes = "$control/smtproutes";

# - - -   Mainline   - - -

foreach $addr (@ARGV) {
	&lookup (0, $addr);
}

exit 0;

#----------------------------------------------------------------- usage ---
sub usage {
	print STDERR <<"EOD_USAGE";
$0 version $version
usage: $0 [options] addr ...
where options are:
	-a uname	set the alias user to 'uname' [$alias]
	-d		enable debugging output
	-e char		set the extension separator to 'char' [$dash]
	-h		show this help
	-q dir		set qmail dir to 'dir' [$qmaildir]
EOD_USAGE
	exit 0;
}

#----------------------------------------------------------------- debug ---
sub debug {
	my ($msg) = @_;

	if ($debug) { print STDERR "DEBUG: $msg\n"; }
0;
}

#----------------------------------------------------------------- show ---
sub show {
	my ($indent, $msg) = @_;

	print +('  ' x $indent) . "$msg\n";
0;
}

#--------------------------------------------------- keep_perl_happy ---
sub keep_perl_happy {
	$opt_d = $opt_h = 0;
};

#--------------------------------------------------- lookup ---
# The core routine.
sub lookup {
	my ($indent, $addr) = @_;
	my ($newaddr, $local, $host, $islocal, $prefix);
	&show($indent, $addr);
	++$indent;

# OK, is this address special?
	if (-f $recipientmap) {
		if ($newaddr = &map_lookup($recipientmap, $addr)) {
			&show($indent, "recipientmap -> $newaddr");
			&lookup($indent+1, $newaddr);
		}
	}

# Split into local and host parts
	($local,$host) = split('@',$addr);

# Apply the various hacks to get a reasonable hostname out of this
	$host = &fixhost($indent, $host);
	return unless (defined $host); # already printed message

# Is it local?

# One of this hosts names?
	if (-f $locals) {
		if ($islocal = &file_lookup($locals,$host)) {
			&show($indent, "- local ($host in control/locals)");
		}
	}

# The official name for this host?
	if (!$islocal && -f $me) {
		if ($islocal = &file_lookup($me,$host)) {
			&show($indent, "- local ($host in contol/me)");
		}
	}

# A virtual-host based here?
	if (!$islocal && -f $virtualdomains) {
		if ($prefix = &map_lookup($virtualdomains,$host)) {
			$islocal = 1;
			&show($indent, 
				"- local ($host in virtualdomains as prefix $prefix)");
		}
	}

# Is it here or somewhere else?
	if ($islocal) {
		&lookup_local($indent,$host,$prefix,$local);
	}

# It's somewhere else
	else {
		&show($indent, "- remote ($host not local)");
		&lookup_remote($indent,$host,$local);
	}
}

#--------------------------------------------------- map_lookup ---
# Look in a old:new file for old, and return new if old was found.
# The map file had better exist or we die here, so check first.
sub map_lookup {
	my ($map, $addr) = @_;
	my ($newaddr, $old, $new);

	open (MAP,"<$map") || die "can't open $map: $!\n";
	while (<MAP>) {
		chomp;
		($old,$new) = split(':',$_);
		if ($old =~ /^$addr$/i) {
			$newaddr = $new;
			last;
		}
	}
	close(MAP);

$newaddr;
}

#---------------------------------------------------- file_lookup ---
sub file_lookup {
	my ($map,$addr) = @_;
	my $found = 0;


	open (LOOKUP,"<$map") || die "can't open $map: $!\n";
	while (<LOOKUP>) {
		chomp;
		if ($_ =~ /^$addr$/i) {
			$found = 1;
			last;
		}
	}
	close(LOOKUP);

$found;
}

#------------------------------------------------------ lookup_local ---
# Look up things that are based here
sub lookup_local {
	my ($indent,$host,$prefix,$local) = @_;
	if ($debug) {
		&debug("lookup_local: passed $indent, $host, $prefix, $local");
	}
	my ($match, $rec, $uid, $homedir, $file, $isalias, $uname);
	my $users = "$qmaildir/users/assign";
	my @temp;

# Try users/assign?
	if (-f "$users") {

# Extract the uid and homedir if we found anything
		if ($rec = &lookup_user($indent, $prefix, $local)) {
			($uname,$uid,$homedir) = (split(':',$rec))[1,2,4];
			if ($uname eq $alias) { $isalias = 1; }
			&show($indent, "- uid=$uid (from users/assign)");
			&show($indent, "- homedir=$homedir (from users/assign)");
		}
	}

# Not using users/assign, or no "+" record => use /etc/passwd
	unless (defined $uid) {
		($uid,$homedir) = (getpwnam($local))[2,7];
		if (defined $uid) {
			&show($indent, "- uid=$uid (from getpwnam($local))");
			&show($indent, "- homedir=$homedir (from getpwnam($local))");
		}
	}

# Try the alias user if ----------------------------------vvvvvvvvvvvvvv
	if (!defined $uid || 				# not found, or
			$uid==0 || 			# found root, or
			!-d $homedir || 		# no homedir, or
			(stat($homedir))[4] != $uid) {	# doesn't own it
		($uid,$homedir) = (getpwnam($alias))[2,7];
		$isalias = 1;
		if (defined $uid) {
			&show($indent, "- uid=$uid from getpwnam($alias)");
			&show($indent, "- homedir=$homedir from getpwnam($alias)");
		}
		else {
			&show($indent, "$local\@$host not found");
			return;
		}
	}

# Pull out the extension, we need it now.
	@temp = split($dash,$local);
	if (defined $isalias) {
		$ext = $local;
	}
	elsif ($#temp > 0) {
		shift @temp; # drop the user
		$ext = join($dash,@temp);
	}
	else { $ext = ''; }
	
# Look for .qmail files
	if ($ext) { $file = "$homedir/.qmail$dash$ext"; }
	else { $file = "$homedir/.qmail"; }

	&debug("local_lookup: looking for $file");
	unless (-f $file) {
		undef $file;
		while ($ext) {
			@temp = split($dash,$ext);
			pop @temp;

			if ($#temp >= 0) {
				$ext = join($dash,@temp);
				$file = "$homedir/.qmail$dash$ext${dash}default";
			}
			else { 
				$ext = '';
				$file = "$homedir/.qmail${dash}default";
			}

			&debug("local_lookup: looking for $file");
			if (-f $file) { last; }
			else  { undef $file};
		}
	}

# Did we find a .qmail file?
	if (defined $file) {
		&show($indent, "- .qmail-file is $file");
		&show_qmail($indent,$file);
	}
	else { &show($indent, "unknown local address"); }
}

#----------------------------------------------------------- show_qmail ---
sub show_qmail {
	my ($indent, $file) = @_;
	my ($first, $rest);

	open ($file, "<$file") || die "can't open $file: $!\n";
	while (<$file>) {
		chomp;
		$first = substr($_,0,1);
		$rest = substr($_,1);

# Pipe to a program?
		if ($first eq '|') { &show($indent, "Program: $rest"); }
# An address?
		elsif ($first eq '&') {
			&show($indent, "Address: $rest");
			&lookup($indent+1, $rest);
		}
# A comment?  No comment.
		elsif ($first eq '#') { }
# A maildir?
		elsif (m"^[./].*/$") { &show($indent, "Maildir: $_"); }
# A mailbox?
		elsif (m"^[./]") { &show($indent, "Mailbox: $_"); }
# A bare address
		else {
			&show($indent, "Address: $_"); 
			&lookup($indent+1, $_);
		}
	}
	close ($file);
}

#------------------------------------------------------ lookup_remote ---
# Look up things that are somewhere else
sub lookup_remote {
	my ($indent,$host,$local) = @_;
	my ($smtphost, $mxhost, $ip);

# Only pull in the resolver if we really need it
	if ($DNS_available) {
		require "Net/DNS.pm";	# DNS-LOOKUP
	}

# Special routing?
	if (-f $smtproutes) {
		if ($smtphost = &map_lookup($smtproutes,$host)) {
			&show($indent, "- smtproutes -> $smtphost");
		}
	}

# Lookup in DNS
	else {

# Found MX record
		if ($mxhost = &lookup_mx($host)) {
			if ($ip = &lookup_a($mxhost)) {
				&show($indent, "remote $mxhost -> $ip (MX)");
			}
			else {
				&show($indent, "remote $mxhost -> ?(dnsfail) (MX)");
			}
		}

# Found A record
		elsif ($ip = &lookup_a($host)) {
			&show($indent, "remote $host -> $ip (A)");
		}

# Nothing found
		else {
			&show($indent, "unresolvable remote $host");
		}
	}
}

#-------------------------------------------------- lookup_user ---
sub lookup_user {
	my ($indent, $prefix, $local) = @_;
	if ($debug) {
		&debug("lookup_user: passed $indent, $prefix, $local");
	}
	my @temp;
	my $sub = '';

# virtual-domains anyone?
	if (defined $prefix) { $local = "$prefix$dash$local"; }

# Exact match
	if ($rec = &lookup_user_record("=$local")) {
		&show($indent, "- exact " .$sub. "user ($rec)");
		return $rec;
	}

# Wildcard match
	while ($local) {
		if ($rec = &lookup_user_record("+$local$dash")) {
			&show($indent, "- wildcard " .$sub. "user ($rec)");
			return $rec;
		}
		@temp = split($dash,$local);
		if ($#temp > 0) {
			pop @temp;
			$local = join($dash, @temp);
		}
		else { last; }
		$sub = 'sub-';
	}

# Generic match
	if ($rec = &lookup_user_record('+')) {
		&show($indent, "- generic user ($rec)");
		return $rec;
	}
undef;
}

#-------------------------------------------------- lookup_user_record ---
sub lookup_user_record {
	my ($userpart) = @_;
	&debug("lookup_user_record: passed $userpart");
	my $key;
	my $file = "$main::qmaildir/users/assign";

	&debug("lookup_user_record: $userpart");
# Load the users file if necessary
	unless (defined %main::users) {
		open (USERS, "<$file") || die "can't open $file: $!\n";
		while (<USERS>) {
			chomp;
			($key) = split(':',$_,2);
			$main::users{$key} = $_;
		}
		close (USERS);
	}

# Actually look for it
	if (defined $main::users{$userpart}) {
		return $main::users{$userpart};
	}
undef;
}

#-------------------------------------------------------------- fromfile ---
# Fetch a value from a file, which must exist.
sub fromfile {
	my ($file) = @_;
	my $value;

	open (FROMFILE, "<$file") || 
		die "can't open $file: $!\n";
	$value = <FROMFILE>;
	chomp $value;
	close(FROMFILE);
$value;
}

#-------------------------------------------------------------- fixhost ---
sub fixhost {
	my ($indent,$host) = @_;

# Do we have a host-part?
	if (!defined $host) {
		if (-f $defaulthost) {
			$host = &fromfile($defaulthost);
			&show($indent, "- host -> $host (from control/defaulthost)");
		}
		elsif (-f $me) {
			$host = &fromfile($me);
			&show($indent, "- host -> $host (from control/me)");
		}
		else {
			&show($indent, "unknown host (and no defaulthost or even me)");
			return undef;
		}
	}

# Unqualified hostname?
	if ($host !~ /\./) {
# This stuff looks wierd, but it *is* what qmail does.
		if (-f $defaultdomain) {
			$host .= &fromfile($defaultdomain);
			&show($indent, "- host -> $host (control/defaultdomain)");
		}
		elsif (-f $me) {
			$host .= &fromfile($me);
			&show($indent, "- host -> $host (control/me)");
		}
		else {
			$host .= 'defaultdomain';
			&show($indent, "- host -> $host (defaultdomain and me missing)");
		}
	}

# Handle the user@host+ case, which I'd never seen before
	if ($host =~ /\+$/) {
		$host = substr($host,0,length($host)-1); # remove plus
		if (-f $plusdomain) {
			$host .= &fromfile($plusdomain);
			&show($indent, "- host -> $host (from control/plusdomain)");
		}
		elsif (-f $me) {
			$host .= &fromfile($me);
			&show($indent, "- host -> $host (from control/me)");
		}
		else {
			$host .= 'plusdomain';
			&show($indent, "- host -> $host (plusdomain and me missing)");
		}
	}
$host;
}

#-------------------------------------------------------------- lookup_mx ---
# Stub out this routine if Net::DNS is unavailable
sub lookup_mx {	#DNS-LOOKUP
	my ($host) = @_;

# Don't just die, 
	return undef unless ($DNS_available);

	my $dns = new Net::DNS::Resolver;
	my (@mx, $mx, $rr, $pref);

	@mx = Net::DNS::mx($dns, $host);

# Got some response
	if (@mx) {
		foreach $rr (@mx) {
			if (defined $mx) {
				if ($rr->preference < $pref) {
					$mx = $rr->exchange;
					$pref = $rr->preference;
					&debug("MX updated -> $mx");
				}
			}
			else {
				$mx = $rr->exchange;
				$pref = $rr->preference;
				&debug("MX found -> $mx");
			}
		}
	}
	else {
		&debug("lookup_mx failed: " . $dns->errorstring);
	}
$mx;
}

#----------------------------------------------------------------- lookup_a ---
# Stub out this routine if Net::DNS is unavailable
sub lookup_a {	#DNS-LOOKUP
	my ($host) = @_;

# Don't just die, say no can do.
	return undef unless ($DNS_available);
	my $dns = new Net::DNS::Resolver;
	my ($query, $rr, $a);

	$query = $dns->search($host);
	if ($query) {
		foreach $rr ($query->answer) {
			next unless $rr->type eq "A";
			$a = $rr->address;
			last;
		}
	}
	else {
		&debug("lookup_a: failed: " . $dns->errorstring);
	}
$a;
}

#------------------------------------------------------------ find_in_inc
# Find a file in the @INC list
sub find_in_inc {
	my ($path) = @_;
	my $prefix;

	foreach $prefix (@INC) {
		if (-f "$prefix/$path") { return 1; }
	}
0;
}


